"""Pytest fixtures for the test suite."""
import pytest
from flask import Flask

from flask_marshmallow import Marshmallow

_app = Flask(__name__)
_app.testing = True


class Bunch(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


# Models


class Author(Bunch):
    pass


class Book(Bunch):
    pass


@pytest.fixture
def mockauthor():
    author = Author(id=123, name="Fred Douglass")
    return author


@pytest.fixture
def mockauthorlist():
    a1 = Author(id=1, name="Alice")
    a2 = Author(id=2, name="Bob")
    a3 = Author(id=3, name="Carol")
    return [a1, a2, a3]


@pytest.fixture
def mockbook(mockauthor):
    book = Book(id=42, author=mockauthor, title="Legend of Bagger Vance")
    return book


@_app.route("/author/<int:id>")
def author(id):
    return "Steven Pressfield"


@_app.route("/authors/")
def authors():
    return "Steven Pressfield, Chuck Paluhniuk"


@_app.route("/books/")
def books():
    return "Legend of Bagger Vance, Fight Club"


@_app.route("/books/<id>")
def book(id):
    return "Legend of Bagger Vance"


@pytest.fixture(scope="function")
def app():
    ctx = _app.test_request_context()
    ctx.push()

    yield _app

    ctx.pop()


@pytest.fixture(scope="function")
def ma(app):
    return Marshmallow(app)


@pytest.fixture
def schemas(ma):
    class AuthorSchema(ma.Schema):
        class Meta:
            fields = ("id", "name", "absolute_url", "links")

        absolute_url = ma.AbsoluteURLFor("author", values={"id": "<id>"})

        links = ma.Hyperlinks(
            {
                "self": ma.URLFor("author", values={"id": "<id>"}),
                "collection": ma.URLFor("authors"),
            }
        )

    class BookSchema(ma.Schema):
        class Meta:
            fields = ("id", "title", "author", "links")

        author = ma.Nested(AuthorSchema)

        links = ma.Hyperlinks(
            {
                "self": ma.URLFor("book", values={"id": "<id>"}),
                "collection": ma.URLFor("books"),
            }
        )

    # So we can access schemas.AuthorSchema, etc.
    return Bunch(**locals())



"""
    flask_marshmallow.validate
    ~~~~~~~~~~~~~~~~~~~~~~~~~~

    Custom validation classes for various types of data.
"""
import os
import re
import typing

from marshmallow.exceptions import ValidationError
from marshmallow.validate import Validator as Validator
from werkzeug.datastructures import FileStorage


def _get_filestorage_size(file: FileStorage) -> int:
    """Return the size of the FileStorage object in bytes."""
    size: int = file.stream.getbuffer().nbytes  # type: ignore
    return size


# This function is copied from loguru with few modifications.
# https://github.com/Delgan/loguru/blob/master/loguru/_string_parsers.py#L35
def _parse_size(size: str) -> float:
    """Return the value which the ``size`` represents in bytes."""
    size = size.strip()
    reg = re.compile(r"([e\+\-\.\d]+)\s*([kmgtpezy])?(i)?(b)", flags=re.I)

    match = reg.fullmatch(size)

    if not match:
        raise ValueError(f"Invalid size value: '{size!r}'")

    s, u, i, b = match.groups()

    try:
        s = float(s)
    except ValueError as e:
        raise ValueError(f"Invalid float value while parsing size: '{s!r}'") from e

    u = "kmgtpezy".index(u.lower()) + 1 if u else 0
    i = 1024 if i else 1000
    b = {"b": 8, "B": 1}[b] if b else 1
    return s * i**u / b


class FileSize(Validator):
    """Validator which succeeds if the file passed to it is within the specified
    size range. If ``min`` is not specified, or is specified as `None`,
    no lower bound exists. If ``max`` is not specified, or is specified as `None`,
    no upper bound exists. The inclusivity of the bounds (if they exist)
    is configurable.
    If ``min_inclusive`` is not specified, or is specified as `True`, then
    the ``min`` bound is included in the range. If ``max_inclusive`` is not specified,
    or is specified as `True`, then the ``max`` bound is included in the range.

    Example: ::

        class ImageSchema(Schema):
            image = File(required=True, validate=FileSize(min='1 MiB', max='2 MiB'))

    :param min: The minimum size (lower bound). If not provided, minimum
        size will not be checked.
    :param max: The maximum size (upper bound). If not provided, maximum
        size will not be checked.
    :param min_inclusive: Whether the ``min`` bound is included in the range.
    :param max_inclusive: Whether the ``max`` bound is included in the range.
    :param error: Error message to raise in case of a validation error.
        Can be interpolated with `{input}`, `{min}` and `{max}`.
    """

    message_min = "Must be {min_op} {{min}}."
    message_max = "Must be {max_op} {{max}}."
    message_all = "Must be {min_op} {{min}} and {max_op} {{max}}."

    message_gte = "greater than or equal to"
    message_gt = "greater than"
    message_lte = "less than or equal to"
    message_lt = "less than"

    def __init__(
        self,
        min: typing.Optional[str] = None,
        max: typing.Optional[str] = None,
        min_inclusive: bool = True,
        max_inclusive: bool = True,
        error: typing.Optional[str] = None,
    ):
        self.min = min
        self.max = max
        self.min_size = _parse_size(self.min) if self.min else None
        self.max_size = _parse_size(self.max) if self.max else None
        self.min_inclusive = min_inclusive
        self.max_inclusive = max_inclusive
        self.error = error

        self.message_min = self.message_min.format(
            min_op=self.message_gte if self.min_inclusive else self.message_gt
        )
        self.message_max = self.message_max.format(
            max_op=self.message_lte if self.max_inclusive else self.message_lt
        )
        self.message_all = self.message_all.format(
            min_op=self.message_gte if self.min_inclusive else self.message_gt,
            max_op=self.message_lte if self.max_inclusive else self.message_lt,
        )

    def _repr_args(self):
        return "min={!r}, max={!r}, min_inclusive={!r}, max_inclusive={!r}".format(
            self.min, self.max, self.min_inclusive, self.max_inclusive
        )

    def _format_error(self, value, message):
        return (self.error or message).format(input=value, min=self.min, max=self.max)

    def __call__(self, value):
        if not isinstance(value, FileStorage):
            raise TypeError(
                f"A FileStorage object is required, not {type(value).__name__!r}"
            )

        file_size = _get_filestorage_size(value)
        if self.min_size is not None and (
            file_size < self.min_size
            if self.min_inclusive
            else file_size <= self.min_size
        ):
            message = self.message_min if self.max is None else self.message_all
            raise ValidationError(self._format_error(value, message))

        if self.max_size is not None and (
            file_size > self.max_size
            if self.max_inclusive
            else file_size >= self.max_size
        ):
            message = self.message_max if self.min is None else self.message_all
            raise ValidationError(self._format_error(value, message))

        return value


class FileType(Validator):
    """Validator which succeeds if the uploaded file is allowed by a given list
        of extensions.

    Example: ::
        class ImageSchema(Schema):
            image = File(required=True, validate=FileType(['.png']))

    :param accept: A sequence of allowed extensions.
    :param error: Error message to raise in case of a validation error.
        Can be interpolated with ``{input}`` and ``{extensions}``.
    """

    default_message = "Not an allowed file type. Allowed file types: [{extensions}]"

    def __init__(
        self,
        accept: typing.Iterable[str],
        error: typing.Optional[str] = None,
    ):
        self.allowed_types = {ext.lower() for ext in accept}
        self.error = error or self.default_message

    def _format_error(self, value):
        return (self.error or self.default_message).format(
            input=value, extensions="".join(self.allowed_types)
        )

    def __call__(self, value):
        if not isinstance(value, FileStorage):
            raise TypeError(
                f"A FileStorage object is required, not {type(value).__name__!r}"
            )

        _, extension = (
            os.path.splitext(value.filename) if value.filename else (None, None)
        )
        if extension is None or extension.lower() not in self.allowed_types:
            raise ValidationError(self._format_error(value))

        return value



import json

from flask import Flask, url_for
from werkzeug.wrappers import Response

from flask_marshmallow import Marshmallow


def test_deferred_initialization():
    app = Flask(__name__)
    m = Marshmallow()
    m.init_app(app)

    assert "flask-marshmallow" in app.extensions


def test_schema(app, schemas, mockauthor):
    s = schemas.AuthorSchema()
    result = s.dump(mockauthor)
    assert result["id"] == mockauthor.id
    assert result["name"] == mockauthor.name
    assert result["absolute_url"] == url_for("author", id=mockauthor.id, _external=True)
    links = result["links"]
    assert links["self"] == url_for("author", id=mockauthor.id)
    assert links["collection"] == url_for("authors")


def test_jsonify_instance(app, schemas, mockauthor):
    s = schemas.AuthorSchema()
    resp = s.jsonify(mockauthor)
    assert isinstance(resp, Response)
    assert resp.content_type == "application/json"
    obj = json.loads(resp.get_data(as_text=True))
    assert isinstance(obj, dict)


def test_jsonify_collection(app, schemas, mockauthorlist):
    s = schemas.AuthorSchema()
    resp = s.jsonify(mockauthorlist, many=True)
    assert isinstance(resp, Response)
    assert resp.content_type == "application/json"
    obj = json.loads(resp.get_data(as_text=True))
    assert isinstance(obj, list)


def test_jsonify_collection_via_schema_attr(app, schemas, mockauthorlist):
    s = schemas.AuthorSchema(many=True)
    resp = s.jsonify(mockauthorlist)
    assert isinstance(resp, Response)
    assert resp.content_type == "application/json"
    obj = json.loads(resp.get_data(as_text=True))
    assert isinstance(obj, list)


def test_links_within_nested_object(app, schemas, mockbook):
    s = schemas.BookSchema()
    result = s.dump(mockbook)
    assert result["title"] == mockbook.title
    author = result["author"]
    assert author["links"]["self"] == url_for("author", id=mockbook.author.id)
    assert author["links"]["collection"] == url_for("authors")



"""
    flask_marshmallow.sqla
    ~~~~~~~~~~~~~~~~~~~~~~

    Integration with Flask-SQLAlchemy and marshmallow-sqlalchemy. Provides
    `SQLAlchemySchema <marshmallow_sqlalchemy.SQLAlchemySchema>` and
    `SQLAlchemyAutoSchema <marshmallow_sqlalchemy.SQLAlchemyAutoSchema>` classes
    that use the scoped session from Flask-SQLAlchemy.
"""
from urllib import parse

import marshmallow_sqlalchemy as msqla
from flask import current_app, url_for
from marshmallow.exceptions import ValidationError

from .schema import Schema


class DummySession:
    """Placeholder session object."""

    pass


class FlaskSQLAlchemyOptsMixin:
    session = DummySession()

    def __init__(self, meta, **kwargs):
        if not hasattr(meta, "sqla_session"):
            meta.sqla_session = self.session
        super().__init__(meta, **kwargs)


# SQLAlchemySchema and SQLAlchemyAutoSchema are available in newer ma-sqla versions
if hasattr(msqla, "SQLAlchemySchema"):

    class SQLAlchemySchemaOpts(FlaskSQLAlchemyOptsMixin, msqla.SQLAlchemySchemaOpts):
        pass

    class SQLAlchemySchema(msqla.SQLAlchemySchema, Schema):
        """SQLAlchemySchema that associates a schema with a model via the
        `model` class Meta option, which should be a
        ``db.Model`` class from `flask_sqlalchemy`. Uses the
        scoped session from Flask-SQLAlchemy by default.

        See `marshmallow_sqlalchemy.SQLAlchemySchema` for more details
        on the `SQLAlchemySchema` API.
        """

        OPTIONS_CLASS = SQLAlchemySchemaOpts

else:
    SQLAlchemySchema = None  # type: ignore

if hasattr(msqla, "SQLAlchemyAutoSchema"):

    class SQLAlchemyAutoSchemaOpts(
        FlaskSQLAlchemyOptsMixin, msqla.SQLAlchemyAutoSchemaOpts
    ):
        pass

    class SQLAlchemyAutoSchema(msqla.SQLAlchemyAutoSchema, Schema):
        """SQLAlchemyAutoSchema that automatically generates marshmallow fields
        from a SQLAlchemy model's or table's column.
        Uses the scoped session from Flask-SQLAlchemy by default.

        See `marshmallow_sqlalchemy.SQLAlchemyAutoSchema` for more details
        on the `SQLAlchemyAutoSchema` API.
        """

        OPTIONS_CLASS = SQLAlchemyAutoSchemaOpts

else:
    SQLAlchemyAutoSchema = None  # type: ignore

auto_field = getattr(msqla, "auto_field", None)


class HyperlinkRelated(msqla.fields.Related):
    """Field that generates hyperlinks to indicate references between models,
    rather than primary keys.

    :param str endpoint: Flask endpoint name for generated hyperlink.
    :param str url_key: The attribute containing the reference's primary
        key. Defaults to "id".
    :param bool external: Set to `True` if absolute URLs should be used,
        instead of relative URLs.
    """

    def __init__(
        self, endpoint: str, url_key: str = "id", external: bool = False, **kwargs
    ):
        super().__init__(**kwargs)
        self.endpoint = endpoint
        self.url_key = url_key
        self.external = external

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        key = super()._serialize(value, attr, obj)
        kwargs = {self.url_key: key}
        return url_for(self.endpoint, _external=self.external, **kwargs)

    def _deserialize(self, value, *args, **kwargs):
        if self.external:
            parsed = parse.urlparse(value)
            value = parsed.path
        endpoint, kwargs = self.adapter.match(value)
        if endpoint != self.endpoint:
            raise ValidationError(
                f'Parsed endpoint "{endpoint}" from URL "{value}"; expected '
                f'"{self.endpoint}"'
            )
        if self.url_key not in kwargs:
            raise ValidationError(
                f'URL pattern "{self.url_key}" not found in {kwargs!r}'
            )
        return super()._deserialize(kwargs[self.url_key], *args, **kwargs)

    @property
    def adapter(self):
        return current_app.url_map.bind("")
